from diagrams import Edge as OGEdge
from random import random
from colorsys import hsv_to_rgb


class Edge:
    @classmethod
    def __makeEdge(self, tooltip=None) -> OGEdge:
        return OGEdge(color=Edge.__getRandomColor(), penwidth="3", tooltip=tooltip)

    @classmethod
    def __getRandomColor(self) -> str:
        r, g, b = [hex(int(255 * i))[-2:] for i in hsv_to_rgb(random(), 0.5, 1.0)]
        return f'#{r}{g}{b}'

    @classmethod
    def connect(self, node1: 'Node', node2: 'Node'):
        node1.obj >> Edge.__makeEdge(f'{node1.getMCName()} -> {node2.getMCName()}') >> node2.obj



from .abstract_node import AbstractNode

class CallManager:

    # Map all callees to all their callers through a manager
    __function_mappings: dict[str, 'CallManager'] = dict()
    __advancement_mappings: dict[str, 'CallManager'] = dict()
    __function_tag_mappings: dict[str, 'CallManager'] = dict()

    def __init__(self):
        # Calls on a per-instance basis
        self.__callers: set['AbstractNode'] = set()

    @staticmethod
    def registerCallToFunction(callee: str, caller: AbstractNode):
        CallManager.__function_mappings.setdefault(callee, CallManager())
        CallManager.__function_mappings[callee].__callers.add(caller)

    @staticmethod
    def registerCallToAdvancement(callee: str, caller: AbstractNode):
        CallManager.__advancement_mappings.setdefault(callee, CallManager())
        CallManager.__advancement_mappings[callee].__callers.add(caller)

    @staticmethod
    def registerCallToFunctionTag(callee: str, caller: AbstractNode):
        CallManager.__function_tag_mappings.setdefault(callee, CallManager())
        CallManager.__function_tag_mappings[callee].__callers.add(caller)

    @staticmethod
    def getCallManagerOfFunction(subject: str):
        return CallManager.__function_mappings.get(subject, CallManager())

    @staticmethod
    def getCallManagerOfAdvancement(subject: str):
        return CallManager.__advancement_mappings.get(subject, CallManager())

    @staticmethod
    def getCallManagerOfFunctionTag(subject: str):
        return CallManager.__function_tag_mappings.get(subject, CallManager())


    @property
    def hasCalls(self):
        return len(self.__callers) > 0

    @property
    def registeredFunctions(self):
        from .function import Function
        for call in self.__callers:
            if isinstance(call, Function):
                yield call

    @property
    def registeredAdvancements(self) -> 'Advancement':
        from .advancement import Advancement
        for call in self.__callers:
            if isinstance(call, Advancement):
                yield call

    @property
    def registeredFunctionTags(self) -> 'FunctionTag':
        from .function_tag import FunctionTag
        for call in self.__callers:
            if isinstance(call, FunctionTag):
                yield call

    @property
    def registeredNodes(self) -> 'Node':
        from .node import Node
        for call in self.__callers:
            if isinstance(call, Node):
                yield call



from .node import Node
from .edge import Edge
from .call_manager import CallManager


class Function(Node):
    function_search_string = r'(?:function )(\S+)'
    advancement_search_string = r'(?:advancement grant .+ (?:only|until|through|from) )(\S+)(?:[$\s])?'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.called_by_tags = []

    @staticmethod
    def getAssociatedGlob() -> str:
        return r'*/functions/**/*.mcfunction'

    @staticmethod
    def handleFile(file: str, base_pack: str):
        return Function(raw_file_path=file, data_folder_path=base_pack)

    def generateCallList(self, file: str):
        # search for function calls
        calls = Node._extractCallListFromFile(file, search_string=Function.function_search_string)
        self._is_recursive = self.getMCName() in calls
        if self._is_recursive:
            calls.remove(self.getMCName())
        for callee in calls:
            CallManager.registerCallToFunction(callee, self)

        # search for advancement grantings
        for call in Node._extractCallListFromFile(file, search_string=Function.advancement_search_string):
            CallManager.registerCallToAdvancement(call, self)

    def initObj(self):
        colors = []
        cm = CallManager.getCallManagerOfFunction(self.getMCName())
        callers = CallManager.getCallManagerOfFunction(self.getMCName())
        if not cm.hasCalls:
            self._is_root_node = True
            colors.append("darkorange2")
        else:
            for caller in cm.registeredFunctionTags:
                self.called_by_tags.append(caller.getMCName())
            if len(self.called_by_tags) > 0:
                # Called by a tag
                colors.append("purple")
        # Single iteration or recursive
        colors.append("firebrick" if self._is_recursive else "royalblue4")
        super()._initObj(colors)

    def createEdges(self):
        cm = CallManager.getCallManagerOfFunction(self.getMCName())
        for caller in cm.registeredNodes:
            Edge.connect(caller, self)

    def _getTooltip(self):
        return super()._getTooltip() + f"\nRecursive: {self._is_recursive}\nCalled by Tags:\n{', '.join(self.called_by_tags)}"



from .node import Node
from .edge import Edge
from .call_manager import CallManager


class Advancement(Node):
    function_search_string = r'(?:"function"\s*:\s*")([^\s"]+)'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def getAssociatedGlob() -> str:
        return r'*/advancements/**/*.json'

    @staticmethod
    def handleFile(file: str, base_pack: str):
        return Advancement(raw_file_path=file, data_folder_path=base_pack)

    def initObj(self):
        super()._initObj(colors = ["limegreen"])

    def generateCallList(self, file: str):
        calls = Node._extractCallListFromFile(file, search_string=Advancement.function_search_string)
        for call in calls:
            CallManager.registerCallToFunction(call, self)

    def createEdges(self):
        cm = CallManager.getCallManagerOfAdvancement(self.getMCName())
        for caller in cm.registeredFunctions:
            Edge.connect(caller, self)
        self._is_root_node = not cm.hasCalls



import re
import os
from abc import ABC, abstractmethod


class AbstractNode(ABC):
    """
        Abstract base class representing a file of a datapack.
        Cannot be drawn to a graph on its own, as not all files need graphical representation.
    """

    def __init__(self, raw_file_path: str, data_folder_path: str):
        self.relative_path = self._getRelativePath(raw_file_path, data_folder_path)
        self._call_name = self._generateCallName(self.relative_path)

        self._display_name = self._generateDisplayName()
        self.generateCallList(raw_file_path)

    @staticmethod
    @abstractmethod
    def handleFile(file: str, base_pack: str):
        raise "Call to abstract base function."

    @abstractmethod
    def generateCallList(self, file: str):
        raise "Call to abstract base function."

    @staticmethod
    @abstractmethod
    def getAssociatedGlob() -> str:
        raise "Call to abstract base function."

    @staticmethod
    def _getRelativePath(file: str, data_folder: str) -> str:
        return os.path.relpath(file, data_folder).replace(os.sep, '/')

    def getMCName(self):
        return self._call_name

    def __eq__(self, other: 'AbstractNode'):
        if not isinstance(other, AbstractNode):
            return NotImplemented
        return type(self) == type(other) and self.getMCName() == other.getMCName()

    def __hash__(self):
        return hash((self.getMCName(), type(self)))

    def _generateDisplayName(self) -> str:
        return self.getMCName().replace(':','/').split('/')[-1]

    @staticmethod
    def _generateCallName(relative_path: str) -> str:
        # Remove Extension
        path = os.path.normpath(relative_path)
        path = os.path.splitext(path)[0]

        # Remove 2nd level of folder structure ("function" or "advancement")
        path = path.replace(os.sep, '/')
        return re.sub('(/[A-z0-9_]+/)', ':', path, count=1)

    @staticmethod
    def _extractCallListFromFile(file: str, search_string: str) -> set[str]:
        calls = set()
        with open(file, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                if not line.startswith("#"):
                    match = re.search(search_string, line)
                    if match is not None:
                        calls.add(match.group(1))
        return set(calls)  # Eliminate duplicates



from .abstract_node import AbstractNode
from .node import Node
from .call_manager import CallManager
import re
import os

class FunctionTag(AbstractNode):
    function_search_string = r'(?:")(\S+:\S+)(?:")'

    def __init__(self, file: str, data_folder: str, *args, **kwargs):
        super().__init__(raw_file_path=file, data_folder_path=data_folder)

    @staticmethod
    def getAssociatedGlob() -> str:
        return r'*/tags/functions/**/*.json'

    @staticmethod
    def handleFile(file: str, base_pack: str):
        FunctionTag(file, base_pack).generateCallList(file)
        return None

    def generateCallList(self, file: str):
        #search for function calls
        calls = AbstractNode._extractCallListFromFile(file, search_string=FunctionTag.function_search_string)
        for callee in calls:
            CallManager.registerCallToFunction(callee, self)

    @staticmethod
    def _generateCallName(relative_path: str) -> str:
        # Remove Extension
        path = os.path.normpath(relative_path)
        path = os.path.splitext(path)[0]

        # Remove 2nd & 3rd level of folder structure ("tags/function")
        path = path.replace(os.sep, '/')
        return "#" + re.sub('(/[A-z0-9_]+/[A-z0-9_]+/)', ':', path, count=1)



from diagrams import Diagram
from glob import iglob, glob
import os
import sys
from .components import Advancement, Cluster, Function, FunctionTag, Node


def generate(data_dirs: list[str], *, output_file: str = 'datapack', format: str = "svg", show: bool = True) -> str:
    '''Generate a diagram of the given datapacks. Returns path to file.'''
    datapacks: list[str] = []
    for dir in data_dirs:
        datapacks += _getDatapackPaths(dir)
    node_list: list[Node] = _makeNodeList(datapacks)
    return _generateGraph(node_list=node_list, output_file=output_file, format=format, show=show)


def _getDatapackPaths(raw_path: str) -> str:
    data_path = ''
    if os.path.basename(raw_path) == "data":
        # Already in /data/ folder
        data_path = raw_path
    else:
        # Assume root folder of datapack
        data_path = os.path.join(raw_path, r'data/')
        if not os.path.exists(data_path):
            # Assume in /datapack/ directory. All relevant data directories should be 1 level deeper.
            data_path = os.path.join(raw_path, r'*/data/')
    return glob(os.path.normpath(data_path))


def _makeNodeList(datapacks: list[str]) -> list[Node]:
    nodes = []
    for pack in datapacks:
        for kind in [Function, Advancement, FunctionTag]:
            for path in iglob(os.path.join(pack, kind.getAssociatedGlob()), recursive=True):
                if artifact := kind.handleFile(file=path, base_pack=pack):
                    nodes.append(artifact)
    if not nodes:
        packs_string = "\n\t".join(datapacks)
        sys.exit(f'Error! There appear to be no files in the datapack. Are you sure this path is correct?\n{packs_string}')
    return nodes


def _establishConnections(node_list: list[Node]):
    for node in node_list:
        node.createEdges()


def _generateGraph(*, output_file: str, node_list: list[Node], format: str, show: bool):
    with Diagram(
        name=output_file,
        graph_attr={
            "bgcolor": "#333333",
            "sep": "100",
            "fontcolor": "white",
            "rankdir": "TB",
            "splines": "ortho",
            "nodesep": "0.3",
            "ranksep": "1",
            "ordering": "in"
        },
        node_attr={
            "fontcolor": "white",
            "fontsize": "15.0",
            "shape": "box",
            "style": "solid",
            "labelloc": "c"
        },
        show=show,
        outformat=format
    ) as d:
        Cluster.clusterize()
        _establishConnections(node_list)
        d.dot = d.dot.unflatten(stagger=3, fanout=True, chain=3)
        return f"{d.filename}.{format}"



#!/usr/bin/env python

# Requires GraphWiz to be installed and added to the system PATH:
# https://graphviz.org/download/

import argparse
import logging
import os
import shlex
import sys

from .datapack_visualizer import generate

# Debug logging toggle
logging.basicConfig(format="", level=logging.DEBUG)


def parse_cli():
    parser = argparse.ArgumentParser(description='Generates a call given one or more datapacks.')
    parser.add_argument('input', metavar='datapack-paths', type=str, nargs='*',
                        help='Path to a specific datapack or a ".../datapacks/" folder.',
                        default=None)
    parser.add_argument('--output', '-o', metavar='output_file', type=str, nargs='?',
                        help='Output file. Defaults to "datapack" if unspecified.',
                        default='datapack')
    parser.add_argument('--format', '-f', metavar='format', type=str, nargs='?',
                        help='Output format (svg/png/pdf/...). "svg" compiles fastest.',
                        default='svg')
    is_posix = os.name == 'posix'
    if len(sys.argv) > 1:
        # Argparse requires comma separated arguments, but to deal with paths including spaces, it has to run through shlex first
        args = parser.parse_args(shlex.split(" ".join(sys.argv[1:]), posix=is_posix))
    else:
        # Argparse requires comma separated arguments, but to deal with paths including spaces, it has to run through shlex first
        args = parser.parse_args(shlex.split(input('Enter the root folder(s) of your datapack(s): '), posix=is_posix))
    # Remove quotation marks from input files
    args.input = [path.strip("'").strip('"') for path in args.input]
    return args


def run():
    args = parse_cli()
    generate(args.input, format=args.format, output_file=args.output)


if __name__ == "__main__":
    run()



from .abstract_node import AbstractNode
from .cluster import Cluster
import re
import os
from diagrams import Node as OGNode
from abc import abstractmethod


class Node(AbstractNode):
    """
        Diagrams-Node wrapper
        Needed because data must be stored before diagrams-Node instance is created, as the context must be set to the correct cluster before init, but that cluster isn't known in advance.
        The actual diagrams-Node creation happens in initObj()
    """

    def __init__(self, raw_file_path: str, data_folder_path: str):
        super().__init__(raw_file_path, data_folder_path)

        self.obj = None  # Node Obj must be created when the appropriate cluster is open
        self._is_root_node = None

        self.parent_cluster_path = '/'.join(self._call_name.replace(':', '/').split('/')[:-1])
        Cluster.fetchClusterByName(self.parent_cluster_path).add_child(self)

    @staticmethod
    def _getRelativePath(file: str, data_folder: str) -> str:
        return os.path.relpath(file, data_folder).replace(os.sep, '/')

    def _getTooltip(self) -> str:
        return self.getMCName() + f"\n\nRoot Node: {self._is_root_node}"

    @abstractmethod
    def initObj(self):
        raise "Call to abstract base function."

    def _initObj(self, colors: list[str]):
        attrs = {
            "width": str(len(self._display_name) / 10.0 + 0.5),
            "height": "0.5",
            "shape": "box",
            "style": "filled",
            "tooltip": self._getTooltip()
        }
        ln = len(colors)
        if ln == 1:
            attrs["fillcolor"] = colors[0]
        if ln == 2:
            # graphviz only allows for 2 colors when using diagonal stripes (╯°□°）╯︵ ┻━┻
            attrs["fillcolor"] = f';{0.5}:'.join(colors)
            attrs["gradientangle"] = "60"
        else:
            weight = 1.0 / (ln * 2)
            attrs["fillcolor"] = f';{weight}:'.join(colors)
            attrs["style"] = "striped"
            attrs["gradientangle"] = "0"
        self.obj = OGNode(label=self._display_name, **attrs)

    @abstractmethod
    def createEdges(self):
        raise "Call to abstract base function."



from diagrams import Cluster as OGCluster
from diagrams import setcluster
import logging


class Cluster:
    __cluster_list: dict[str, 'Cluster'] = {}

    def __init__(self, name: str):
        self.__name = name
        self._display_name = name.split('/')[-1]
        self.__cluster_obj: OGCluster = None
        self.__cluster_list.setdefault(self.__name, self)
        self.__children: list['Node'] = []
        self.__depth = name.count('/')

    def add_child(self, child: 'Node'):
        self.__children.append(child)

    @classmethod
    def fetchClusterByName(self, name: str) -> 'Cluster':
        # Fetch cluster if already created, otherwise create a new instnace.
        return Cluster.__cluster_list.get(name, Cluster.createTreeRec(name, name.split('/')))

    @classmethod
    def createTreeRec(self, own_name: str, path: list[str]):
        # If parent cluster already exists, create a new cluster.
        # If parent does not exist, create parent recursively until root cluster.
        own_name = '/'.join(path)
        parent_cluster = '/'.join(path[:-1])
        if len(path) > 1 and not parent_cluster in Cluster.__cluster_list:
            Cluster.createTreeRec(parent_cluster, path[:-1])
        if not own_name in Cluster.__cluster_list:
            ret = Cluster(own_name)
            Cluster.__cluster_list.setdefault(own_name, ret)
            return ret

    def clusterDepthToColor(self, iteration) -> str:
        return f'#{str((iteration%4)+1)*6}'

    def enterCluster(self):
        setcluster(self.__cluster_obj)
        logging.debug(f'-> {self.__name}')

    def leaveCluster(self):
        if self.__cluster_obj._parent:
            self.__cluster_obj._parent.subgraph(self.__cluster_obj.dot)
        else:
            self.__cluster_obj._diagram.subgraph(self.__cluster_obj.dot)
        setcluster(self.__cluster_obj._parent)
        logging.debug(f'<- {self.__name}')

    def draw(self) -> OGCluster:
        logging.debug(f'[+] cluster: {self.__name} Depth: {self.__depth}')
        attrs = {
            "bgcolor": self.clusterDepthToColor(self.__depth),
            "fontsize": "22",
            "fontcolor": "white",
            "style": "solid",
            "tooltip": self.__name
        }
        self.__cluster_obj = OGCluster(label=self._display_name, graph_attr=attrs)

    @classmethod
    def clusterize(self):
        open_clusters_stack: list[Cluster] = []

        for name, cluster in sorted(Cluster.__cluster_list.items()):
            cluster: Cluster
            while len(open_clusters_stack) > 0 and not name.startswith(open_clusters_stack[-1].__name + '/'):
                open_clusters_stack.pop(-1).leaveCluster()
            cluster.draw()
            cluster.enterCluster()
            open_clusters_stack.append(cluster)
            for node in cluster.__children:
                if node.obj is None:
                    node.initObj()
        while len(open_clusters_stack) > 0:
            open_clusters_stack.pop(-1).leaveCluster()
